//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_PREFERENCES_HH
#define BMP_PREFERENCES_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif //HAVE_CONFIG_H

#include <glibmm.h>
#include <gtkmm.h>
#include <libglademm.h>
#include <vector>
#include <map>
#include <set>
#include <mcs/mcs.h>

struct BmpTooltipEntry
{
    char *widget_name;
    char *tooltip_text;
    char *stock_id;
};

namespace Bmp
{
  /** Preferences dialog
   *
   * Bmp::Preferences is a complex dialog for adjusting run time parameters
   * of BMP trough the GUI instead of having to manipulate the configuration
   * file.
   */
  class Preferences : public Gtk::Window
  {

  public:

      Preferences (BaseObjectType                        *cobject,
                   const Glib::RefPtr<Gnome::Glade::Xml> &xml);
      static Preferences* create ();
      ~Preferences () {}

      enum AudioSink
      {
          SINK_ALSA,
          SINK_OSS,
          SINK_SUNAUDIO,
          SINK_ESD,
          SINK_HAL,
      };

  private:

      void on_category_changed ();

      class CategoryView;
      class ListColumnView;
#ifdef HAVE_HAL
      class HALView;
#endif //HAVE_HAL

      Glib::RefPtr<Gnome::Glade::Xml>     m_ref_xml;

      Gtk::Notebook                     * notebook_categories;
      CategoryView                      * category_view;

      /* Audio */
#ifdef HAVE_ALSA
      struct AlsaDevice
      {
        /** Description of the device
        *
        */
        std::string   description;

        /** ALSA device string of this device
        *
        */
        std::string   dev;
            
        /** ALSA card id of this device 
        *
        */
        int	          card_id;

        /** ALSA device id of this device
        *
        */
        int	          device_id;
      };
      typedef std::vector<AlsaDevice> AlsaDevices;

      struct AlsaCard
      {
        /** Description of the card
         *
         */
        std::string   description;

        /** ALSA device string of this card 
         *
         */
        std::string   dev;

        /** ALSA card id of this card 
         *
         */
        int	          card_id;

        /** List of devices on this card 
         *
         */
        AlsaDevices   devices;  
    };
    typedef std::vector<AlsaCard> AlsaCards;

    // ALSA card columns record
    class AlsaCardColumnRecord : public Gtk::TreeModel::ColumnRecord
    {
        public:

          Gtk::TreeModelColumn<Glib::ustring> name;
          Gtk::TreeModelColumn<AlsaCard>      card;

          AlsaCardColumnRecord ()
          {
              add (name);
              add (card);
          }
    };

    // ALSA device columns record
    class AlsaDeviceColumnRecord : public Gtk::TreeModel::ColumnRecord
    {
      public:

        Gtk::TreeModelColumn<Glib::ustring> name;
        Gtk::TreeModelColumn<AlsaDevice>    device;

        AlsaDeviceColumnRecord ()
        {
            add (name);
            add (device);
        }
    };

    AlsaCardColumnRecord     alsa_card_columns;
    AlsaDeviceColumnRecord   alsa_device_columns;

    /** Returns a list of all ALSA cards present on the system, each containing a list of devices
     *
     * @returns A list of ALSA cards
     *
     */
    AlsaCards get_alsa_cards ();

    void on_cbox_alsa_card_changed ();

    Gtk::ComboBox                     * cbox_alsa_card;
    Gtk::ComboBox                     * cbox_alsa_device;
    Gtk::SpinButton                   * alsa_buffer_time;
    Glib::RefPtr<Gtk::ListStore>        list_store_alsa_cards;
    Glib::RefPtr<Gtk::ListStore>        list_store_alsa_device;

#endif //HAVE_ALSA

    // Audio System column record
    class AudioSystemColumnRecord : public Gtk::TreeModel::ColumnRecord
    {
      public:

        Gtk::TreeModelColumn<Glib::ustring> description;
        Gtk::TreeModelColumn<std::string>   sink_name;
        Gtk::TreeModelColumn<int>           tab_id;
        Gtk::TreeModelColumn<AudioSink>     sink_id;

        AudioSystemColumnRecord()
        {
            add (description);
            add (sink_name);
            add (tab_id);
            add (sink_id);
        }
    };
    AudioSystemColumnRecord  audio_system_columns;

    void setup_audio ();

    void audio_system_apply_set_sensitive ();

    void on_cbox_audio_system_changed ();
    void on_audio_system_apply ();

    Gtk::Button                       * audio_system_apply;
    Gtk::ComboBox                     * cbox_audio_system;
    Gtk::HBox                         * audio_system_changed_warning;
    Gtk::Notebook                     * notebook_audio_system;
    Glib::RefPtr<Gtk::ListStore>        list_store_audio_systems;
    std::set<Glib::ustring>             m_sinks;

    // OSS
    Gtk::ComboBoxEntry                * cbe_oss_device;
    Gtk::SpinButton                   * oss_buffer_time;

    // ESD
    Gtk::Entry                        * esd_host;
    Gtk::SpinButton                   * esd_buffer_time;

#ifdef HAVE_SUN
    Gtk::ComboBoxEntry                * cbe_sun_device;
    Gtk::SpinButton                   * sun_buffer_time;
#endif //HAVE_SUN

    /* Podcasts */
    Gtk::FileChooserButton            * m_podcasts_fcb;

#ifdef HAVE_HAL
    /* HAL Devices View */
    HALView                           * hal_view;
#endif //HAVE_HAL

  }; // class Preferences
} // namespace Bmp

#endif // BMP_PREFERENCES_HH
